/* Low level interface for debugging NetApp threads for GDB, the GNU debugger.
   $Id: //depot/prod/freebsd/R5.3/usr/src/contrib/gdb/gdb/ntap-thread.c#1 $

This file is part of GDB.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/* This module implements a sort of half target that sits between the
   machine-independent parts of GDB and the remote interface (remote.c) to
   provide access to the NetApp thread implementation.  */

#include "defs.h"

#include "gdbthread.h"
#include "target.h"
#include "inferior.h"
#include "bfd.h"    /* BFD_HOST_64BIT_LONG */
#include "gdbcmd.h"		/* define "set debug na_unwind" */
#include <string.h>
#include <assert.h>

static int unwind_debug = 1;
#define DD(fmt,args...)  (unwind_debug ? fprintf_unfiltered (gdb_stdlog, "   > %s: " fmt, __FUNCTION__, args) :0)

extern struct target_ops netapp_thread_ops; /* Forward declaration */
extern struct target_ops netapp_core_ops; /* Forward declaration */

/* place to store core_ops before we overwrite it */
static struct target_ops orig_core_ops;

struct target_ops netapp_thread_ops;
struct target_ops netapp_core_ops;

extern struct target_ops remote_ops; /* target vector for remote.c */
extern struct target_ops core_ops; /* target vector for corelow.c */

/* TRUE if we've already scanned for threads. */
static int scanned_for_threads;

/* Address of the "sk" process table. */
static CORE_ADDR sk_proc_table;

/* Size of an entry in the "sk" process table. */
static int sk_proc_size;

/* PID of the currently-running process, which is the index of the
   currently running process in the process table - we can't use
   the address as a PID, as PIDs are "int"s in GDB and a pointer
   won't fit in an "int" on Alpha.

   The indices are 1-origin, as a PID of 0 seems to mean something
   special to GDB. */
int sk_running_proc_pid;

/* Macros to convert PIDs to pointers to the process, and *vice versa*. */
#define	PID_TO_PROC_PTR(pid)	(sk_proc_table + ((pid) - 1)*sk_proc_size)
#define	PROC_PTR_TO_PID(proc_ptr) \
	(((proc_ptr) - sk_proc_table)/sk_proc_size + 1)

static int netapp_get_sk_nprocs (void);
static void netapp_get_sk_proc_size (void);
static CORE_ADDR netapp_get_sk_proc_table (void);
static int netapp_is_sk_proc (int pid);
static int netapp_get_sk_running_proc_pid ();
static void fetch_thread_registers (int regno);
static void fetch_thread_register (CORE_ADDR sp, int regno);
static CORE_ADDR netapp_get_proc_sp (ptid_t ptid);
void netapp_subclass_core_ops (struct target_ops *p);


/* Return 1 if thread 'tid' was active.  Overrides the default
   core_file_thread_alive() from corelow.c which glibly always
   returns 1: but then the common operation 'thread apply all bt'
   stops on the first dead thread because the backtrace command
   usually gets a memory error
*/
static int
netapp_core_file_thread_alive (ptid_t ptid)
{
    int   pid = PIDGET(ptid);
    int   isa_proc = netapp_is_sk_proc(pid);
    return isa_proc;
}

/* Convert an "sk" process ID to a string.  Returns the string in a static
   buffer.  */

static char *
netapp_pid_to_str (ptid_t ptid)
{
  static char buf[40];
  int   pid = PIDGET(ptid);

  if (netapp_is_sk_proc(pid)) {
      if (pid == sk_running_proc_pid)
	  sprintf (buf, "sk crashed process %d", pid-1);
      else
	  sprintf (buf, "sk process %d", pid-1);
  }
  else
      sprintf (buf, "DEAD PROCESS PLACEHOLDER");

  return buf;
}

static void
netapp_thread_detach (args, from_tty)
     char *args;
     int from_tty;
{
  unpush_target (&netapp_thread_ops);
  remote_ops.to_detach (args, from_tty);
}

/* A PID is just a pointer to a "struct sk_Proc".  Its context is
   in the "context" member of that structure. */

/* A context switch is done by calling "sk_hw_switch()".

   On the 386, "sk_hw_switch()" does a context switch is done by:

	Pushing %ebp
	Pushing %edi
	Pushing %esi
	Pushing %ebx
	Storing %esp into the "sp" member of the "context" member of
	  the "sk_Proc" structure for the process.

   When "sk_hw_switch" is called, its first argument is a pointer
   to the "context" member of the process giving up control, i.e.
   a pointer to the place where the SP of that process will be
   stored.

   Other registers are treated as caller-save registers, so we don't
   care what values that have, which is good because they're not saved
   across a context switch.

   Thus, what we have is:

	SP for process ->   EBX for process
			    ESI for process
			    EDI for process
			    EBP for process
			    EIP of caller of "sk_hw_switch()
			    pointer to ESP of caller of "sk_hw_switch()"

   as the EIP was pushed by the "call" and the pointer to the ESP of
   the caller is the argument to "sk_hw_switch()" pushed before the
   call, so we fetch 6 pointer's worth of data pointed to by the SP,
   and get the five registers in question from that, and the ESP from
   the location pointed to by the 6th value.

   This will return a set of registers that make it look as if
   "sk_hw_switch()" had just returned; I guess if this really
   confuses people we can hack it to back the EIP up so that it looks
   as if we're just about to call "sk_hw_switch()".  It seems kind of
   pointless to leave "sk_hw_switch()" on the stack, though, unless
   not doing so screws up our ability to do certain things to threads
   other than the currently active one. */

/* Mappings from tm-i386v.h */
#define EBX_IN_CONTEXT	0
#define ESI_IN_CONTEXT	1
#define EDI_IN_CONTEXT	2
#define EBP_IN_CONTEXT	3
#define EIP_IN_CONTEXT	4
#define ESP_IN_CONTEXT	5
#define NOT_IN_CONTEXT	-1	/* not available for other threads */

//static int regmap[I386_NUM_GREGS] = {
static int regmap[NUM_GREGS] = {
  NOT_IN_CONTEXT,		/* EAX */
  NOT_IN_CONTEXT,		/* ECX */
  NOT_IN_CONTEXT,		/* EDX */
  EBX_IN_CONTEXT,
  ESP_IN_CONTEXT,
  EBP_IN_CONTEXT,
  ESI_IN_CONTEXT,
  EDI_IN_CONTEXT,
  EIP_IN_CONTEXT,
  NOT_IN_CONTEXT,
  NOT_IN_CONTEXT,		/* XXX - CS; need to get common value? */
  NOT_IN_CONTEXT,		/* XXX - SS; need to get common value? */
  NOT_IN_CONTEXT,		/* XXX - DS; need to get common value? */
  NOT_IN_CONTEXT,		/* XXX - ES; need to get common value? */
  NOT_IN_CONTEXT,		/* FS */
  NOT_IN_CONTEXT,		/* GS */
};

/* Register numbers of various important registers. */

#define EAX_REGNUM	0
#define ECX_REGNUM	1
#define EDX_REGNUM	2
#define EBX_REGNUM	3
#define ESP_REGNUM	4
#define EBP_REGNUM	5
#define ESI_REGNUM	6
#define EDI_REGNUM	7
#define EIP_REGNUM	8


/* Fetch registers for "sk" process whose PID is "inferior_pid".
   Must not be called if "inferior_pid" equals "sk_running_proc",
   as the registers for the running process are not where they'd
   be if the thread weren't running. */

static void
fetch_thread_register (CORE_ADDR sp, int regno)
{
  int i;
  CORE_ADDR regaddr;
  char buf[TARGET_PTR_BIT / TARGET_CHAR_BIT];
  CORE_ADDR spp;

  i = regmap[regno];
  if (i == NOT_IN_CONTEXT)
    return;			/* we can't fetch it */
  regaddr = sp + i*REGISTER_RAW_SIZE(regno);
  if (regno == ESP_REGNUM)
    {
      /* Special case - the context contains a pointer to the SP value,
         not the SP value itself.
         XXX - if it returns a non-zero value, we should fail! */
      target_read_memory (regaddr, buf, TARGET_PTR_BIT / TARGET_CHAR_BIT);
      spp = extract_address (buf, TARGET_PTR_BIT / TARGET_CHAR_BIT);
      target_read_memory (spp, buf, REGISTER_RAW_SIZE(regno));
    }
  else
    {
      /* Get the register at the offset from the context.
         XXX - if it returns a non-zero value, we should fail! */
      target_read_memory (regaddr, buf, REGISTER_RAW_SIZE(regno));
    }

  supply_register (regno, buf);
}


static void
netapp_thread_fetch_registers (regno)
     int regno;
{
  struct cleanup *old_chain;
  int i;

  if (inferior_ptid.pid && inferior_ptid.pid != sk_running_proc_pid)
    {
      /* We're not in the currently-running thread, so fetch the
         registers using that thread's context and stack. */

	CORE_ADDR sp;
	
	/* Get the SP for the process. */
	sp = netapp_get_proc_sp (inferior_ptid);
	
	if (regno >= 0) {
	    /* Get the register in question. */
	    fetch_thread_register (sp, regno);
	}
	else {
	    /* Get all the registers. */
	    for (regno = 0; regno < NUM_REGS; regno++)
		fetch_thread_register (sp, regno);
	}
    }
  else
    {
      /* Save the current inferior PID, and set it to 0 so that we don't
	 try to fetch the registers from some per-thread register section,
	 because we don't have per-thread register sections.  */
      old_chain = save_inferior_ptid ();
      inferior_ptid.pid = 0;
      if (target_has_execution)
	remote_ops.to_fetch_registers (regno);
      else
	orig_core_ops.to_fetch_registers (regno);
      do_cleanups (old_chain);
    }    	
}


int
netapp_thread_maybe_fetch_registers (int regno)
{
    if (!(inferior_ptid.pid && inferior_ptid.pid != sk_running_proc_pid))
	return 0;
    netapp_thread_fetch_registers (regno);
    return 1;
}

static void
netapp_thread_store_registers (regno)
     int regno;
{
  struct cleanup *old_chain;
  char regs[REGISTER_BYTES];

  if (inferior_ptid.pid && inferior_ptid.pid != sk_running_proc_pid)
    {
      /* We're not in the currently-running thread, so fetch the
         registers using that thread's context and stack. */
#if 0
      store_thread_registers (regno, regs);
#else
      abort();	/* XXX - implement me */
#endif
    }
  else
    {
      /* Save the current inferior PID, and set it to 0 so that we don't
	 try to fetch the registers from some per-thread register section,
	 because we don't have per-thread register sections.  */
      old_chain = save_inferior_ptid ();
      inferior_ptid.pid = 0;
      if (target_has_execution)
	remote_ops.to_fetch_registers (regno);
      else
	orig_core_ops.to_fetch_registers (regno);
      do_cleanups (old_chain);
    }    	
}


static int
symtab_offsetof (struct type *sym_type, char *fieldname, struct type **field_type)
{
    struct symbol *sym;
    int    i, offset;

    /* Look for the offsets of "context" and "state" in an "sk_Proc". */
    for (i = TYPE_NFIELDS (sym_type)-1; i>=TYPE_N_BASECLASSES (sym_type); i--){

	if (strcmp(TYPE_FIELD_NAME (sym_type, i), fieldname) == 0) {

	    /* Found member.  Is it some C++ "static member"? */
	    if (TYPE_FIELD_STATIC (sym_type, i)) return -3;

	    /* Its offset into the structure in bits
	       is TYPE_FIELD_BITPOS (sk_Proc_type, i).
	       Compute its offset in bytes. */
	    offset = TYPE_FIELD_BITPOS (sym_type, i)/8;

	    if (field_type) {
		/* Also, remember its type, so we can find the offset of the
		   "sp" member of that structure. */
		*field_type = check_typedef (TYPE_FIELD_TYPE (sym_type, i));
		
		/* Make sure it *is* a structure. */
		if (TYPE_CODE (*field_type) != TYPE_CODE_STRUCT) return -2;
	    }

	    return offset;
	}
    }
    return -4;
}



/* Build a thread list from the "sk" process table. */

static void
netapp_find_new_threads ()
{
  int i;
  int sk_nprocs;
  ptid_t ptid;

  /* XXX - for now, don't do this on live appliances. */
  if (target_has_execution)
    return;

  /* If we already got the PID of the currently running process, we've
     already fetched the threads, and, as we're looking at a core dump,
     once we've fetched them once, there's no need to fetch them again. */
  if (scanned_for_threads)
    return;
  scanned_for_threads = 1;

  /* Get the size of the process table. */
  sk_nprocs = netapp_get_sk_nprocs ();

  DD ("sk_nprocs %d\n", sk_nprocs);

  /* Get the address of the process table. */
  sk_proc_table = netapp_get_sk_proc_table ();
  if (!sk_proc_table)
    {
      /* Failed to fetch the address, or it's NULL because it
         hasn't been allocated yet.
         Can't fetch the list of threads. */
      return;
    }

  /* Get the size of an entry in the process table. */
  netapp_get_sk_proc_size ();
  if (sk_proc_size == 0)
    {
      /* Failed to find the size of a process table entry.
         Can't fetch the list of threads. */
      return;
    }

  /* Set sk_nprocs to be the last active thread (skipping unused
     threads at the end.)
     Then create a gdb thread for every process in [0..sk_nprocs)
     whether or not it is active, so that gdb thread ids match sk_proc 
     indices.  The cost of making numbers match up is it exposes dead
     threads to the user.  They can do "thread 1053" and get junk, and
     maybe think the corefile was incomplete or there is a gdb bug.  If
             p sk_global.gl_proc_table[1052].state
     shows SK_PROC_UNUSED (index 1052 for thread 1053) it's user error.
  */
  while (sk_nprocs > 0 && ! netapp_is_sk_proc(sk_nprocs))
    {
      sk_nprocs--;
    }

  /* Loop through the process table looking for live processes, and make
     a thread for each one. */
  for (i = 0; i < sk_nprocs; i++)
    {
      ptid.pid = i + 1;	/* PIDs start at 1 */
      /* Make a thread. */
      add_thread (ptid);
    }

  /* Set "inferior_pid" to the PID of the currently running process. */
  inferior_ptid.pid = netapp_get_sk_running_proc_pid ();
}


/*
  Return the PID of the sk process which abended.

  This PID's registers are fetched from the core register state
  instead of sk_proc_table[PID].context.sp, by checking for it
  in fetch_thread_registers ().
*/

static int
netapp_get_sk_running_proc_pid ()
{
  struct minimal_symbol *sk_running_proc_symp;
  static CORE_ADDR sk_running_proc_addr;
  CORE_ADDR sk_running_proc;

  /* Get the address of "sk_running_proc", if we don't already have it. */
  if (sk_running_proc_addr == 0)
    {
      sk_running_proc_symp = lookup_minimal_symbol ("sk_running_proc",
	NULL, (struct objfile *) NULL);
      if (sk_running_proc_symp == NULL)
        return 0;	/* "should not happen" - return error here? */
      sk_running_proc_addr = SYMBOL_VALUE_ADDRESS (sk_running_proc_symp);
    }

  /* Get the value of "sk_running_proc".
     XXX - if it returns a non-zero value, we should fail! */
  sk_running_proc = read_memory_unsigned_integer (sk_running_proc_addr, 4);

  /*
    Temporary code because sk_running_proc is not defined in the trap
    crash path of freebsd yet.  It *is* defined if sk_panic() is
    called from an assertion check.  So: if the value is 0 then assume
    CPU 0's thread abended and obtain the PID from
    sk_processor_ptrs[0]->pr_running_proc.
  */
  if (sk_running_proc == 0) {
      int n;
      struct minimal_symbol *skpp;
      CORE_ADDR   skpp0, skpp_addr;
      struct type *sym_type;
      
      skpp = lookup_minimal_symbol ("sk_processor_ptrs",
              NULL, (struct objfile *) NULL);
      if (skpp == NULL)
        return 0;	/* "should not happen" - return error here? */

      skpp_addr = SYMBOL_VALUE_ADDRESS (skpp);
      skpp0 = read_memory_unsigned_integer (skpp_addr + 0*4, 4);

      sym_type = lookup_transparent_type ("sk_processor");
      if (!sym_type) return 0;
      n = symtab_offsetof (sym_type, "pr_running_proc", 0);

      sk_running_proc = read_memory_unsigned_integer (skpp0 + n, 4);
  }

  /* Now convert it to a PID. */
  sk_running_proc_pid = PROC_PTR_TO_PID(sk_running_proc);

  DD ("sk_running_proc_pid %d\n", sk_running_proc_pid);
  return sk_running_proc_pid;
}


/* Get the size of the process table.
   XXX - assumes an "int" on the netapp fits in an "int" on the
   machine debugging the netapp. */
static int
netapp_get_sk_nprocs ()
{
  struct minimal_symbol *sk_nprocs_symp;
  CORE_ADDR sk_nprocs_addr;
  int sk_nprocs;

  /* Get the address of "sk_nprocs". */
  sk_nprocs_symp = lookup_minimal_symbol ("sk_nprocs", NULL,
	(struct objfile *) NULL);
  if (sk_nprocs_symp == NULL)
    {
      /* Symbol isn't defined; we have an older netapp software
         release that doesn't save the size of the process table
	 in the core dump.
	 
	 Use the value from the old #define.
	 
	 XXX - it went from 600 in 1.0 to 700 in 2.0, so we may
	 think it's bigger than it is on old core dumps. */
      sk_nprocs = /*1000*/700;	/* XXX - make this 700 again eventually,
      			   once we no longer have 3.0 cores that don't
			   have "sk_nprocs" */
    }
  else
    {
      sk_nprocs_addr = SYMBOL_VALUE_ADDRESS (sk_nprocs_symp);

      /* Get the value of "sk_nprocs".
	 XXX - if it returns a non-zero value, we should fail! */
      sk_nprocs = read_memory_unsigned_integer (sk_nprocs_addr, 4);
    }
  return sk_nprocs;
}

/* Offset, in bytes, of the "sp" member of the "context" member of an
   "sk_Proc", relative to the beginning of an "sk_Proc". */
static int sp_offset;

/* Offset, in bytes, of the "state" member of an "sk_Proc", relative to
   the beginning of an "sk_Proc". */
static int state_offset;

/* Values of "SK_PROC_UNUSED" and "SK_PROC_IS_MSG" members of an
   "sk_Ptype" - and flag indicating whether there *is* an SK_PROC_IS_MSG
   member. */
static int sk_proc_unused;
static int sk_proc_is_msg;
static int sk_proc_is_msg_found;

/*
  Get info about ontap process structs via gdb symbol table queries:
      sk_proc_size : sizeof (sk_Proc)
      sp_offset : offsetof (sk_Proc, context.sp)
      state_offset : offsetof (sk_Proc, state)
      sk_proc_unused : SK_PROC_UNUSED
*/

static void
netapp_get_sk_proc_size ()
{
  struct symbol *sk_Proc_sym, *sk_Ptype_sym;
  register struct type *sk_Proc_type, *offset_type = 0, *sk_Ptype_type;
  int i;
  int context_found = 0;
  int state_found = 0;
  int sk_proc_unused_found = 0;

  sk_Proc_sym = lookup_symbol ("sk_Proc", (struct block *) NULL,
		       VAR_NAMESPACE, 0, (struct symtab **) NULL);
  if (sk_Proc_sym == NULL)
    {
      warning ("sk_Proc type not found; can't find sk processes.");
      return;
    }
  sk_Proc_type = check_typedef (SYMBOL_TYPE (sk_Proc_sym));
  if (TYPE_CODE (sk_Proc_type) != TYPE_CODE_STRUCT)
    {
      warning ("sk_Proc is not a struct; can't find sk processes.");
      return;
    }

  /* Look for the offsets of "context" and "state" in an "sk_Proc". */
  for (i = TYPE_NFIELDS (sk_Proc_type) - 1;
		i >= TYPE_N_BASECLASSES (sk_Proc_type); i--)
    {
      if (strcmp(TYPE_FIELD_NAME (sk_Proc_type, i), "context") == 0)
	{
	  /* Found the "context" member.  Is it some C++ "static member"? */
	  if (TYPE_FIELD_STATIC (sk_Proc_type, i))
	    {
	      warning ("\"context\" member of sk_Proc is weird; can't find sk processes.");
	      return;
	    }

	  /* Its offset, in bits, from the beginning of an "sk_Proc"
	     is TYPE_FIELD_BITPOS (sk_Proc_type, i).
	     Compute its offset in bytes. */
	  sp_offset = TYPE_FIELD_BITPOS (sk_Proc_type, i)/8;

	  /* Also, remember its type, so we can find the offset of the
	     "sp" member of that structure. */
	  offset_type = check_typedef (TYPE_FIELD_TYPE (sk_Proc_type, i));

	  /* Make sure it *is* a structure. */
	  if (TYPE_CODE (offset_type) != TYPE_CODE_STRUCT)
	    {
	      warning ("\"context\" member of sk_Proc is not a struct; can't find sk processes.");
	      return;
	    }
	  context_found = 1;
	}
      else if (strcmp(TYPE_FIELD_NAME (sk_Proc_type, i), "state") == 0)
	{
	  /* Found the "state" member.  Is it some C++ "static member"? */
	  if (TYPE_FIELD_STATIC (sk_Proc_type, i))
	    {
	      warning ("\"state\" member of sk_Proc is weird; can't find sk processes.");
	      return;
	    }

	  /* Its offset, in bits, from the beginning of an "sk_Proc"
	     is TYPE_FIELD_BITPOS (sk_Proc_type, i).
	     Compute its offset in bytes. */
	  state_offset = TYPE_FIELD_BITPOS (sk_Proc_type, i)/8;

	  /* XXX - remember the type and size of "state"? */
	  state_found = 1;
	}
    }

  if (!context_found)
    {
      /* We didn't find a "context" member. */
      warning ("sk_Proc has no \"context\" member; can't find sk processes.");
      return;
    }

  if (!state_found)
    {
      /* We didn't find a "state" member. */
      warning ("sk_Proc has no \"state\" member; can't find sk processes.");
      return;
    }

  /* Find the offset of the "sp" member of "context". */
  for (i = TYPE_NFIELDS (offset_type) - 1;
		i >= TYPE_N_BASECLASSES (offset_type); i--)
    {
      if (strcmp(TYPE_FIELD_NAME (offset_type, i), "sp") == 0)
	{
	  /* Found the "sp" member.  Is it some C++ "static member"? */
	  if (TYPE_FIELD_STATIC (offset_type, i))
	    {
	      warning ("\"sp\" member of sk_Proc.context is weird; can't find sk processes.");
	      return;
	    }

	  /* Its offset, in bits, from the beginning of "context"
	     is TYPE_FIELD_BITPOS (offset_type, i).
	     Compute its offset in bytes, and add it to "sp_offset". */
	  sp_offset += TYPE_FIELD_BITPOS (offset_type, i)/8;

	  /* XXX - remember the size of "sp"? */
	  goto found;
	}
    }

    /* We didn't find an "sp" member. */
    warning ("sk_Proc.context has no \"sp\" member; can't find sk processes.");
    return;

found:
  /* Now find the values of the "SK_PROC_UNUSED" and "SK_PROC_IS_MSG"
     members of "sk_Pstate", which is an "enum". */
  sk_Ptype_sym = lookup_symbol ("sk_Pstate", (struct block *) NULL,
		       VAR_NAMESPACE, 0, (struct symtab **) NULL);
  if (sk_Ptype_sym == NULL)
    {
      warning ("sk_Ptype type not found; can't find sk processes.");
      return;
    }
  sk_Ptype_type = check_typedef (SYMBOL_TYPE (sk_Ptype_sym));
  if (TYPE_CODE (sk_Ptype_type) != TYPE_CODE_ENUM)
    {
      warning ("sk_Ptype is not an enum; can't find sk processes.");
      return;
    }

  /* Look for the values of "SK_PROC_UNUSED" and "SK_PROC_IS_MSG" in
     "sk_Ptype". */
  for (i = TYPE_NFIELDS (sk_Ptype_type) - 1; i >= 0; i--)
    {
      if (strcmp(TYPE_FIELD_NAME (sk_Ptype_type, i), "SK_PROC_UNUSED") == 0)
	{
	  /* Its value, as an integral value, is
	     TYPE_FIELD_BITPOS (sk_Ptype_type, i). */
	  sk_proc_unused = TYPE_FIELD_BITPOS (sk_Ptype_type, i);
	  sk_proc_unused_found = 1;
	}
      else if (strcmp(TYPE_FIELD_NAME (sk_Ptype_type, i), "SK_PROC_IS_MSG") == 0)
	{
	  /* Its value, as an integral value, is
	     TYPE_FIELD_BITPOS (sk_Ptype_type, i). */
	  sk_proc_is_msg = TYPE_FIELD_BITPOS (sk_Ptype_type, i);
	  sk_proc_is_msg_found = 1;
	}
    }

  if (!sk_proc_unused_found)
    {
      /* We didn't find an "SK_PROC_UNUSED" member. */
      warning ("sk_Ptype has no \"SK_PROC_UNUSED\" member; can't find sk processes.");
      return;
    }

  /* Set "sk_proc_size" to the length of an "sk_Proc". */
  sk_proc_size = TYPE_LENGTH (SYMBOL_TYPE (sk_Proc_sym));
}


#if 1
/* Get the address of "sk_proc_table".  On success, returns 1 and sets
   "sk_proc_table"; on failure, returns 0. */
static CORE_ADDR
netapp_get_sk_proc_table ()
{
  struct minimal_symbol *sk_proc_table_symp;
  CORE_ADDR sk_proc_table_addr;

  /* Get the address of "sk_proc_table". */
  sk_proc_table_symp = lookup_minimal_symbol ("gdb_sk_proc_table", NULL,
	(struct objfile *) NULL);
  if (sk_proc_table_symp == NULL)
    return 0;		/* not defined? */

  sk_proc_table_addr = SYMBOL_VALUE_ADDRESS (sk_proc_table_symp);

  /* Get the value of "sk_proc_table".
     XXX - if it returns a non-zero value, we should fail! */
  return read_memory_integer (sk_proc_table_addr, 4);
}

#else

/*
  Get the address of "sk_proc_table".  On success, returns 1 and sets
  "sk_proc_table"; on failure, returns 0. 
*/

static CORE_ADDR
netapp_get_sk_proc_table ()
{
    int i;
    struct symbol *skg_sym;
    struct type *skg_type;
    CORE_ADDR  sk_proc_table;
    
    /* Get the address of "sk_global.gl_proc_table". */

    skg_sym = lookup_symbol ("sk_global", (struct block *) NULL,
		       VAR_NAMESPACE, 0, (struct symtab **) NULL);
    if (skg_sym == NULL) 	goto error;
    skg_type = SYMBOL_TYPE (skg_sym)->target_type;
    if (!skg_type || TYPE_CODE (skg_type) != TYPE_CODE_STRUCT)	goto error;

    sk_proc_table = 0;
    for (i = TYPE_NFIELDS (skg_type)-1; i>=TYPE_N_BASECLASSES (skg_type); i--){

	if (strcmp (TYPE_FIELD_NAME (skg_type, i), "gl_proc_table") == 0) {

	    /* Is it some C++ "static member"? */
	    if (TYPE_FIELD_STATIC (skg_type, i)) goto error;
	    
	    /* Its offset into the structure, in bits
	       is TYPE_FIELD_BITPOS (skg_type, i).
	       Compute its offset in bytes. */
	    sk_proc_table = TYPE_FIELD_BITPOS (skg_type, i) / 8 +
		SYMBOL_VALUE_ADDRESS (skg_sym);
	    sk_proc_table = read_memory_unsigned_integer (sk_proc_table, 4);
	    break;
	}
    }
    DD ("sk_proc_table at 0x%s\n", paddr_nz (sk_proc_table));
    return sk_proc_table;

 error:
    warning ("sk_global has no \"gl_proc_table\" member; can't find sk processes.");
    return 0;
}
#endif


/*
  Return TRUE if there's an "sk" process with the PID "pid". 
*/

static int
netapp_is_sk_proc (int pid)
{
  int state;
  CORE_ADDR proc_ptr;
  
  /* Find the address of the process. */
  proc_ptr = PID_TO_PROC_PTR (pid);

  /* Fetch the "state" member of the "sk_Proc" for the process in
     question. */
  state = read_memory_integer (proc_ptr + state_offset, 4);

  /* There's a process there if the state is neither SK_PROC_UNUSED
     or SK_PROC_IS_MSG (assuming there *is* an SK_PROC_IS_MSG state). */
  return ! (state == sk_proc_unused
	    || (sk_proc_is_msg_found && state == sk_proc_is_msg));
}

/*
  Get the SP for a given "sk" process with the PID "pid". 
*/

static CORE_ADDR
netapp_get_proc_sp (ptid_t ptid)
{
  int pid = ptid_get_pid(ptid);
  CORE_ADDR proc_ptr;

  /* Find the address of the process. */
  proc_ptr = PID_TO_PROC_PTR(pid);
  
  /* Fetch the "sp" member "context" member of the "sk_Proc" for the
     process in question.
     XXX - if it returns a non-zero value, we should fail! */
  return read_memory_integer (proc_ptr + sp_offset, 4);
}

static void
netapp_core_detach (args, from_tty)
     char *args;
     int from_tty;
{
  unpush_target (&core_ops);
  orig_core_ops.to_detach (args, from_tty);
}

/* Fork an inferior process, and start debugging it with /proc.  */

static void
netapp_thread_create_inferior (exec_file, allargs, env)
     char *exec_file;
     char *allargs;
     char **env;
{
  remote_ops.to_create_inferior (exec_file, allargs, env);

  push_target (&netapp_thread_ops);
}

/* Clean up after the inferior dies.  */

static void
netapp_thread_mourn_inferior ()
{
  unpush_target (&netapp_thread_ops);
  remote_ops.to_mourn_inferior ();
}

static void
netapp_thread_core_detach (args, from_tty)
     char *args;
     int from_tty;
{
  unpush_target (&core_ops);
  orig_core_ops.to_detach (args, from_tty);
}

static void
init_netapp_thread_ops ()
{
  netapp_thread_ops = remote_ops;
  netapp_thread_ops.to_shortname = "netapp-threads";
  netapp_thread_ops.to_longname = "Netapp threads.";
  netapp_thread_ops.to_doc = "NetApp thread support.";
  netapp_thread_ops.to_detach = netapp_thread_detach;
  netapp_thread_ops.to_create_inferior = netapp_thread_create_inferior;
  netapp_thread_ops.to_mourn_inferior = netapp_thread_mourn_inferior;
  netapp_thread_ops.to_fetch_registers = netapp_thread_fetch_registers;
  netapp_thread_ops.to_store_registers = netapp_thread_store_registers;
  netapp_thread_ops.to_pid_to_str = netapp_pid_to_str;
  netapp_thread_ops.to_find_new_threads = netapp_find_new_threads;
}

static void
init_netapp_core_ops ()
{
  netapp_core_ops = core_ops;
  netapp_core_ops.to_shortname = "netapp-core";
  netapp_core_ops.to_longname = "NetApp core threads.";
  netapp_core_ops.to_doc = "NetApp thread support for core files.";
  netapp_core_ops.to_detach = netapp_thread_core_detach;
  netapp_core_ops.to_fetch_registers = netapp_thread_fetch_registers;
  netapp_core_ops.to_pid_to_str = netapp_pid_to_str;
  netapp_core_ops.to_find_new_threads = netapp_find_new_threads;
  netapp_core_ops.to_thread_alive = netapp_core_file_thread_alive;
}

void
netapp_subclass_core_ops (struct target_ops *p)
{
    p->to_pid_to_str = netapp_pid_to_str;
    p->to_find_new_threads = netapp_find_new_threads;
    p->to_thread_alive = netapp_core_file_thread_alive;
}

/* we suppress the call to add_target of core_ops in corelow because
   if there are two targets in the stratum core_stratum, find_core_target
   won't know which one to return.  see corelow.c for an additonal
   comment on coreops_suppress_target. */
//int coreops_suppress_target = 1;

void
_initialize_ntap_thread ()
{
#if 0
  init_netapp_thread_ops();
  init_netapp_core_ops();

  add_target (&netapp_thread_ops);

  memcpy(&orig_core_ops, &core_ops, sizeof (struct target_ops));
  memcpy(&core_ops, &netapp_core_ops, sizeof (struct target_ops));
  add_target (&core_ops);
#endif

  add_show_from_set (add_set_cmd ("na_unwind", no_class, var_zinteger,
                                  (char *) &unwind_debug,
                                  "Set debugging of netapp frame unwind.\n",
                                  &setdebuglist),
                     &showdebuglist);

}
